#include <stdlib.h>
#include <math.h>
#include "terrain.h"

#include <stdio.h>

// generates a bunch of ridges, containing n nodes at a radial separation dr,
// having heights h +/- dh; then convolves same with profile
typedef struct
{
	float x, y, h;
} ridgenode;

#define URD	((float)rand() / (float)RAND_MAX)

void trn_ridges(terrain_t *map, terrain_t *profile, int n, float dr, float h, float dh)
{
	ridgenode *nodes;
	char *connectivity;
	int i, j, k;
	float s, t;
	float r, theta;

	// XXX
	float min, max;

	if (!map) return;

	// map must be power-of-two dimensioned
	if (((map->w - 1) | map->w) >> 1 != (map->w - 1) ||
	    ((map->h - 1) | map->h) >> 1 != (map->h - 1))
		return;
		
	nodes = (ridgenode*)malloc(n * sizeof(ridgenode));
	if (!nodes) return;

	connectivity = malloc(n * n);
	if (!connectivity)
	{
		free(nodes);
		return;
	}

	memset(connectivity, 0, n * n);

	// generate ridge nodes
	for (i = 0; i < n; i++)
	{
		r = i * dr;
		theta = URD * TWOPI;
		nodes[i].x = r * cos(theta);
		nodes[i].y = r * sin(theta);
		nodes[i].h = h + 2 * (URD - 0.5) * dh;
	}
	
	// now, link up the ridge nodes by joining each one to its two nearest
	// unlinked neighbours
#define S(a)	((a)*(a))
	for (i = 0; i < n; i++)
	{
		for (j = 1, k = j - 1, s = 10 * n * n * dr * dr; j < n; j++)
		if (i != j && !connectivity[i + j * n] &&
			(t = S(nodes[i].x - nodes[j].x) +
				S(nodes[i].y - nodes[j].y)) < s)
		{
			s = t;
			k = j;
		}

		trn_drawline(map, nodes[i].x, nodes[i].y, nodes[i].h,
				nodes[k].x, nodes[k].y, nodes[k].h);

		connectivity[i + k * n]++;
		connectivity[k + i * n]++;
	}

	// make the convolution of the map with the profile given
	trn_convolve(map, profile);
}

// add roughness to the landscape, with amplitude mag * (1 + alpha * h)
// and standard deviation k
void trn_addroughness(terrain_t *map, float mag, float k, float alpha)
{
	terrain_t *noise;
	float *a, *b;
	float min, max;

	if (!map) return;

	// map must be power-of-two dimensioned
	if (((map->w - 1) | map->w) >> 1 != (map->w - 1) ||
	    ((map->h - 1) | map->h) >> 1 != (map->h - 1))
		return;

	noise = trn_new(map->w, map->h, map->xs); // assume isotropy

	if (!noise)
		return;

	trn_gaussiannoise(noise, mag, k);

	for (a = map->terrain, min = max = *a; a < map->terrain + map->w * map->h; a++)
	{
		if (*a > max) max = *a;
		if (*a < min) min = *a;
	}

	printf("%f %f\n", min, max);

	for (a = map->terrain, b = noise->terrain; a < (map->terrain + map->w * map->h); a++, b++)
		*a += *b * (1 + alpha * *a);

	trn_free(noise);
}
